%INCLUDE "common.asm"

;=============================================================================
; Constants
;=============================================================================

SECTION .rodata:
ALIGN 16

AVS_WORD_LOW_MASK    : TIMES 4 DW 0x00FF
AVS_WORD_HIGH_MASK   : TIMES 4 DW 0xFF00

;=============================================================================
; Load / separate luma macro
;=============================================================================

;load two lines (%9, %9+%10), extract luma in (%1,%2), extract packed chroma in (%3,%4)
%MACRO LOAD_SEPARATE_LUMA_TWO_LINES 10
   movq        %1, [%9]                   ; load new lines
   movq        %5, [%9 + 8]
   add         %9, %10
   movq        %2, [%9]
   movq        %6, [%9 + 8]
   sub         %9, %10
   
   movq        %3, %1                     ; copy
   movq        %7, %5
   movq        %4, %2
   movq        %8, %6
   
   pand        %1, [AVS_WORD_LOW_MASK]    ; separate luma
   pand        %2, [AVS_WORD_LOW_MASK]      
   pand        %5, [AVS_WORD_LOW_MASK]
   pand        %6, [AVS_WORD_LOW_MASK]

   pand        %3, [AVS_WORD_HIGH_MASK]   ; separate chroma
   pand        %7, [AVS_WORD_HIGH_MASK]
   pand        %4, [AVS_WORD_HIGH_MASK]
   pand        %8, [AVS_WORD_HIGH_MASK]
   
   psrlw       %3, 8                      ; chroma
   psrlw       %7, 8
   psrlw       %4, 8
   psrlw       %8, 8
   
   packuswb    %1, %5                     ; luma
   packuswb    %2, %6                     ; luma
   packuswb    %3, %7                     ; U1|V1|U2|V2|U3|V3|U4|V4
   packuswb    %4, %8                     ; U1|V1|U2|V2|U3|V3|U4|V4
%ENDMACRO

;=============================================================================
; Separate chroma macro
;=============================================================================

; take two packed chroma line in (%1,%2), output unpacked chroma on 16 bytes line in (%3,%4) (fst) and (%1,%2) (snd)
%MACRO SEPARATE_CHROMA_TWO_LINES 4
   movq        %3, %1
   movq        %4, %2
   
   pand        %1, [AVS_WORD_HIGH_MASK]
   pand        %2, [AVS_WORD_HIGH_MASK]
   
   pand        %3, [AVS_WORD_LOW_MASK]
   pand        %4, [AVS_WORD_LOW_MASK]
   
   psrlw       %1, 8
   psrlw       %2, 8
%ENDMACRO

;=============================================================================
; Interpolation macro
;=============================================================================

; %1,%3,%5,%7 contains four lines of chroma 1, %2,%4,%6,%8 contains four lines of chroma 2
; %9, %10, %11, %12, %13 and %14 contains the coefficients of the interpolation
; outputs in %1 ( chroma 1 ), %2 ( chroma 2 ) the interpolation 
%MACRO INTERPOLATION       14
   AVS_MULTIPLY_ %+ %9     %1
   AVS_MULTIPLY_ %+ %9     %2
   AVS_MULTIPLY_ %+ %10    %3
   AVS_MULTIPLY_ %+ %10    %4
   AVS_MULTIPLY_ %+ %11    %5
   AVS_MULTIPLY_ %+ %11    %6
   AVS_MULTIPLY_ %+ %12    %7
   AVS_MULTIPLY_ %+ %12    %8
   paddw       %1, %3
   paddw       %2, %4
   paddw       %5, %7
   paddw       %6, %8
   paddw       %1, %5
   paddw       %2, %6
   paddw       %1, [AVS_WORD_ %+ %13]
   paddw       %2, [AVS_WORD_ %+ %13]
   psrlw       %1, %14
   psrlw       %2, %14
   packuswb    %1, %1
   packuswb    %2, %2
%ENDMACRO

;=============================================================================
; Conversion macro ( 2x8 pixels )
;=============================================================================
   
; %1 != 0      --> first line
; %2 != 0      --> last line
; %3,%3+%4     --> two lines to read
; %5,%5+%6     --> two luma lines to write
; %7,%8        --> chroma lines to write
; %9,%10       --> temporary buffer
; %11          --> interpolation macros
; %12, %13, %14, %15, %16, %17 --> coefficients
%MACRO PROCESS_TWO_LINES 17

   LOAD_SEPARATE_LUMA_TWO_LINES mm0, mm1, mm2, mm3, mm4, mm5, mm6, mm7, %3, %4

%IF %1 == 0
   movq        [%5], mm0                 ; write luma
%ENDIF
%IF %2 == 0
   add         %5, %6
   movq        [%5], mm1
   sub         %5, %6
%ENDIF

   SEPARATE_CHROMA_TWO_LINES mm2, mm3, mm4, mm5

%IF %1 == 0
   movq        mm6, [%9]                 ; load previous chroma 1
   movq        mm7, [%10]
   movq        mm0, [%9 + 8]             ; load previous chroma 2
   movq        mm1, [%10 + 8]
%ENDIF   
   
%IF %2 == 0
   movq        [%9], mm4                 ; save current chroma 1
   movq        [%10], mm5
   movq        [%9 + 8], mm2             ; save current chroma 2
   movq        [%10 + 8], mm3
%ENDIF   
   
%IF %1 == 0
   %11         mm6, mm0, mm7, mm1, mm4, mm2, mm5, mm3, %12, %13, %14, %15, %16, %17
   movd        [%7], mm6                 ; write chroma 1
   movd        [%8], mm0                 ; write chroma 1
%ENDIF
   
%ENDMACRO

;=============================================================================
; Conversion macro ( frame )
;=============================================================================

; %1 --> name
; %2 --> conversion macro
; %3, %4, %5, %6, %7, %8 --> coefficients   
%MACRO CONVERSION_FRAME 8   

   avs_mangle yuy2toyv12_frame_ %+ %1 %+ _isse

yuy2toyv12_frame_ %+ %1 %+ _isse:
   AVS_SET_STACK     3, esi, edi, ebp, ebx
   
   .pDstY            PARAM(1)
   .nDstYPitch       PARAM(2)
   .pDstU            PARAM(3)
   .nDstUPitch       PARAM(4)
   .pDstV            PARAM(5)
   .nDstVPitch       PARAM(6)
   .pSrc             PARAM(7)
   .nSrcPitch        PARAM(8)
   .nWidth           PARAM(9)
   .nHeight          PARAM(10)
   
   .nLoopCounter     PSTACK(0)
   .nWidthDiv2       PSTACK(1)
   .nWidthMul2       PSTACK(2)
   
   mov               ebp, esp
   
   ; allocates the two chroma temporary lines
   sub               esp, [ebp + .nWidth]
   sub               esp, [ebp + .nWidth]
   mov               edx, esp
   sub               esp, [ebp + .nWidth]
   sub               esp, [ebp + .nWidth]
   
   mov               esi, [ebp + .pSrc]
   mov               edi, [ebp + .pDstY]
   mov               ebx, [ebp + .pDstU]
   mov               ecx, [ebp + .pDstV]
   
   mov               eax, [ebp + .nWidth]
   shl               eax, 1
   mov               [ebp + .nWidthMul2], eax
   shr               eax, 2
   mov               [ebp + .nWidthDiv2], eax
   
   xor               eax, eax

.x_first_loop

   PROCESS_TWO_LINES 1, 0, esi, 0, edi, 0, ebx, ecx, esp + eax, edx + eax, %2, %3, %4, %5, %6, %7, %8
   
   add               eax, 16
   add               esi, 16
   add               edi, 8
   add               ebx, 4
   add               ecx, 4
   
   cmp               eax, [ebp + .nWidthMul2]
   jnz               .x_first_loop
   
   add               esi, [ebp + .nSrcPitch]
   add               edi, [ebp + .nDstYPitch]
   sub               esi, [ebp + .nWidthMul2]
   sub               edi, [ebp + .nWidth]
   sub               ebx, [ebp + .nWidthDiv2]
   sub               ecx, [ebp + .nWidthDiv2]
   
   mov               eax, [ebp + .nHeight]
   sub               eax, 2                     ; first and last lines
   shr               eax, 1
   mov               [ebp + .nLoopCounter], eax
   
.y_loop

   xor               eax, eax

.x_loop

   PROCESS_TWO_LINES 0, 0, esi, [ebp + .nSrcPitch], edi, [ebp + .nDstYPitch], ebx, ecx, esp + eax, edx + eax, %2, %3, %4, %5, %6, %7, %8
   
   add               eax, 16
   add               esi, 16
   add               edi, 8
   add               ebx, 4
   add               ecx, 4
   
   cmp               eax, [ebp + .nWidthMul2]
   jnz               .x_loop
   
   add               esi, [ebp + .nSrcPitch]
   add               edi, [ebp + .nDstYPitch]
   add               ebx, [ebp + .nDstUPitch]
   add               ecx, [ebp + .nDstVPitch]
   sub               esi, [ebp + .nWidthMul2]
   sub               edi, [ebp + .nWidth]
   sub               ebx, [ebp + .nWidthDiv2]
   sub               ecx, [ebp + .nWidthDiv2]
   add               esi, [ebp + .nSrcPitch]
   add               edi, [ebp + .nDstYPitch]

   mov               eax, [ebp + .nLoopCounter]
   dec               eax
   mov               [ebp + .nLoopCounter], eax
   jnz               .y_loop
   
   xor               eax, eax
   
.x_last_loop

   PROCESS_TWO_LINES 0, 1, esi, 0, edi, 0, ebx, ecx, esp + eax, edx + eax, %2, %3, %4, %5, %6, %7, %8
   
   add               eax, 16
   add               esi, 16
   add               edi, 8
   add               ebx, 4
   add               ecx, 4
   
   cmp               eax, [ebp + .nWidthMul2]
   jnz               .x_last_loop
   
   add               esp, [ebp + .nWidthMul2]
   add               esp, [ebp + .nWidthMul2]
   
   AVS_REL_STACK
   
%ENDMACRO   

;=============================================================================
; Code
;=============================================================================

SECTION .text:

CONVERSION_FRAME 1, INTERPOLATION, 1, 3, 3, 1, 4, 3
CONVERSION_FRAME 2, INTERPOLATION, 1, 2, 1, 0, 2, 2
CONVERSION_FRAME 3, INTERPOLATION, 0, 1, 2, 1, 2, 2
CONVERSION_FRAME 1_top, INTERPOLATION, 3, 7, 5, 1, 8, 4
CONVERSION_FRAME 1_bottom, INTERPOLATION, 1, 5, 7, 3, 8, 4
