#include "conversion.h"
#include "yuy2toyv12.h"

#define KERNEL(A, B, C, D, x1, x2, x3, x4) (( (x1) * (A) + (x2) * (B) + (x3) * (C) + (x4) * (D) + 8 ) >> 4)

#define KERNEL_2_6_6_2(x1, x2, x3, x4) KERNEL( 2, 6, 6, 2, x1, x2, x3, x4 )
#define KERNEL_1_5_7_3(x1, x2, x3, x4) KERNEL( 1, 5, 7, 3, x1, x2, x3, x4 )
#define KERNEL_3_7_5_1(x1, x2, x3, x4) KERNEL( 3, 7, 5, 1, x1, x2, x3, x4 )
#define KERNEL_1_2_1_0(x1, x2, x3, x4) (( (x1) + (x2) * 2 + (x3) + 2 ) >> 2)
#define KERNEL_0_1_2_1(x1, x2, x3, x4) KERNEL_1_2_1_0( x4, x3, x2, x1 )

#define DECLARE_LINE_FUNCTION(kernel, name) \
static void yuy2toyv12_line_ ## name(uint8_t *pDstY1, uint8_t *pDstY2, uint8_t *pDstU, uint8_t *pDstV, \
                                     const uint8_t *pSrc1, const uint8_t *pSrc2, \
                                     const uint8_t *pSrc3, const uint8_t *pSrc4, \
                                     int nWidth) \
{ \
   int x; \
\
   for ( x = 0; x < nWidth >> 1; x++ ) \
   { \
      pDstY1[x*2] = pSrc2[x*4]; \
      pDstY1[x*2+1] = pSrc2[x*4+2]; \
      pDstY2[x*2] = pSrc3[x*4]; \
      pDstY2[x*2+1] = pSrc3[x*4+2]; \
\
      pDstU[x] = KERNEL_ ## kernel( pSrc1[x*4+1], pSrc2[x*4+1], pSrc3[x*4+1], pSrc4[x*4+1] ); \
      pDstV[x] = KERNEL_ ## kernel( pSrc1[x*4+3], pSrc2[x*4+3], pSrc3[x*4+3], pSrc4[x*4+3] ); \
   } \
}

#define DECLARE_FRAME_FUNCTION(name, kernel) \
void yuy2toyv12_frame_ ## name(uint8_t *pDstY, int nDstYPitch, \
                               uint8_t *pDstU, int nDstUPitch, \
                               uint8_t *pDstV, int nDstVPitch, \
                               const uint8_t *pSrc, int nSrcPitch, \
                               int nWidth, int nHeight) \
{ \
   int y; \
\
   yuy2toyv12_line_ ## kernel( pDstY, pDstY + nDstYPitch, pDstU, pDstV, \
                               pSrc, pSrc, pSrc + nSrcPitch, pSrc + 2 * nSrcPitch, \
                               nWidth ); \
\
   pDstY += nDstYPitch * 2; \
   pSrc += nSrcPitch * 2; \
   pDstU += nDstUPitch; \
   pDstV += nDstVPitch; \
\
   for ( y = 1; y < (nHeight >> 1) - 1; y++ ) \
   { \
      yuy2toyv12_line_ ## kernel( pDstY, pDstY + nDstYPitch, pDstU, pDstV, \
                                  pSrc - nSrcPitch, pSrc, pSrc + nSrcPitch, pSrc + 2 * nSrcPitch, \
                                  nWidth ); \
\
      pDstY += nDstYPitch * 2; \
      pSrc += nSrcPitch * 2; \
      pDstU += nDstUPitch; \
      pDstV += nDstVPitch; \
   } \
\
   yuy2toyv12_line_ ## kernel( pDstY, pDstY + nDstYPitch, pDstU, pDstV, \
                               pSrc - nSrcPitch, pSrc, pSrc + nSrcPitch, pSrc + nSrcPitch, \
                               nWidth ); \
}

#define DECLARE_FIELD_FUNCTION(name, name_top, name_bottom) \
void yuy2toyv12_field_ ## name(uint8_t *pDstY, int nDstYPitch, \
                               uint8_t *pDstU, int nDstUPitch, \
                               uint8_t *pDstV, int nDstVPitch, \
                               const uint8_t *pSrc, int nSrcPitch, \
                               int nWidth, int nHeight) \
{ \
   yuy2toyv12_frame_ ## name_top( pDstY, nDstYPitch * 2, pDstU, nDstUPitch * 2, \
                                  pDstV, nDstVPitch * 2, pSrc, nSrcPitch * 2, \
                                  nWidth, nHeight / 2 ); \
   yuy2toyv12_frame_ ## name_bottom( pDstY + nDstYPitch, nDstYPitch * 2, pDstU + nDstUPitch, nDstUPitch * 2, \
                                     pDstV + nDstVPitch, nDstVPitch * 2, pSrc + nSrcPitch, nSrcPitch * 2, \
                                     nWidth, nHeight / 2 ); \
}

DECLARE_LINE_FUNCTION( 2_6_6_2, 1_3_3_1 );
DECLARE_LINE_FUNCTION( 1_5_7_3, 1_5_7_3 );
DECLARE_LINE_FUNCTION( 3_7_5_1, 3_7_5_1 );
DECLARE_LINE_FUNCTION( 0_1_2_1, 0_1_2_1 );
DECLARE_LINE_FUNCTION( 1_2_1_0, 1_2_1_0 );

DECLARE_FRAME_FUNCTION( 1, 1_3_3_1 );
DECLARE_FRAME_FUNCTION( 1_top, 3_7_5_1 );
DECLARE_FRAME_FUNCTION( 1_bottom, 1_5_7_3 );
DECLARE_FIELD_FUNCTION( 1, 1_top, 1_bottom );

DECLARE_FRAME_FUNCTION( 2, 1_2_1_0 );
DECLARE_FRAME_FUNCTION( 3, 0_1_2_1 );
DECLARE_FIELD_FUNCTION( 2, 2, 1 );
DECLARE_FIELD_FUNCTION( 3, 1, 3 );

extern yuy2toyv12_f yuy2toyv12_frame_1_top_isse;
extern yuy2toyv12_f yuy2toyv12_frame_1_bottom_isse;
DECLARE_FIELD_FUNCTION( 1_isse, 1_top_isse, 1_bottom_isse );
DECLARE_FIELD_FUNCTION( 2_isse, 2_isse, 1_isse );
DECLARE_FIELD_FUNCTION( 3_isse, 1_isse, 3_isse );
