%INCLUDE "common.asm"

;=============================================================================
; Interpolation macro ( generic )
;=============================================================================

; unpacks %1, %2, %3 and %4, using %9 which is NULL
; computes ( %1 * %5 + %3 * %6 + %7 ) >> %8 in %1
; computes ( %2 * %5 + %4 * %6 + %7 ) >> %8 in %2
; %5 and %6 must not be null
%MACRO INTERPOLATION 9
   punpcklbw   %1, %9
   punpcklbw   %2, %9
   punpcklbw   %3, %9
   punpcklbw   %4, %9
   
   AVS_MULTIPLY_ %+ %5 %1
   AVS_MULTIPLY_ %+ %5 %2
   AVS_MULTIPLY_ %+ %6 %3
   AVS_MULTIPLY_ %+ %6 %4
   
   paddw       %1, %3
   paddw       %2, %4
   paddw       %1, [AVS_WORD_ %+ %7]
   paddw       %2, [AVS_WORD_ %+ %7]
   psraw       %1, %8
   psraw       %2, %8
%ENDMACRO

;=============================================================================
; Conversion macros ( generic, copy first, copy second, average )
;=============================================================================

; %1 is a luma row
; %2, %3 are chroma 1 and 2 first rows
; %4, %5 are chroma 1 and 2 second rows
; %6 is the destination row
; %7, %8, %9, %10 are the interpolation coefficients
; mm7 is a blank register
%MACRO CONVERSION_INTERPOLATION 10
   movq        mm0, [ %1 ]
   movd        mm2, [ %2 ]
   movd        mm3, [ %3 ]
   movd        mm4, [ %4 ]
   movd        mm5, [ %5 ]
   movq        mm1, mm0
   
   INTERPOLATION mm2, mm3, mm4, mm5, %7, %8, %9, %10, mm7
   
   packuswb    mm2, mm2
   packuswb    mm3, mm3
   
   punpcklbw   mm2, mm3                      ; UVUVUVUV
   punpcklbw   mm0, mm2                      ; YUYVYUYV
   punpckhbw   mm1, mm2                      ; YUYVYUYV
   
   movq        [ %6 ], mm0
   movq        [ %6 + 8 ], mm1
%ENDMACRO

%MACRO CONVERSION_COPY_FIRST     10
   movq        mm0, [ %1 ]
   movd        mm2, [ %2 ]
   movd        mm3, [ %3 ]
   movq        mm1, mm0
   
   punpcklbw   mm2, mm3                      ; UVUVUVUV
   punpcklbw   mm0, mm2                      ; YUYVYUYV
   punpckhbw   mm1, mm2                      ; YUYVYUYV

   movq        [ %6 ], mm0
   movq        [ %6 + 8 ], mm1
%ENDMACRO

%MACRO CONVERSION_COPY_SECOND    10   
   movq        mm0, [ %1 ]
   movd        mm2, [ %4 ]
   movd        mm3, [ %5 ]
   movq        mm1, mm0
   
   punpcklbw   mm2, mm3                      ; UVUVUVUV
   punpcklbw   mm0, mm2                      ; YUYVYUYV
   punpckhbw   mm1, mm2                      ; YUYVYUYV

   movq        [ %6 ], mm0
   movq        [ %6 + 8 ], mm1
%ENDMACRO   

%MACRO CONVERSION_AVERAGE        10
   movq        mm0, [ %1 ]
   movd        mm2, [ %2 ]
   movd        mm3, [ %3 ]
   movd        mm4, [ %4 ]
   movd        mm5, [ %5 ]
   movq        mm1, mm0
   
   pavgb       mm2, mm4
   pavgb       mm3, mm5
   
   punpcklbw   mm2, mm3                      ; UVUVUVUV
   punpcklbw   mm0, mm2                      ; YUYVYUYV
   punpckhbw   mm1, mm2                      ; YUYVYUYV
   
   movq        [ %6 ], mm0
   movq        [ %6 + 8 ], mm1
%ENDMACRO

;=============================================================================
; Conversion macro ( whole line )
;=============================================================================

; %1 is the luma row pointer
; %2, %3 are the chroma 1 and 2 first row pointers
; %4, %5 are the chroma 1 and 2 second row pointers
; %13 is the destination row pointer
; %6 is the interpolation macro to use
; %7 is the loop counter register to use
; %8 is it's initial value
; %9, %10, %11, %12 are the interpolation coefficients 
%MACRO CONVERSION_LINE           13
   mov               %7, [ %8 ]
%%loop_label
   sub               %7, 4
   %6                %1 + 2 * %7, %2 + %7, %3 + %7, %4 + %7, %5 + %7, %13 + 4 * %7, %9, %10, %11, %12
   jnz               %%loop_label
%ENDMACRO

;=============================================================================
; Conversion macro ( whole frame )
;=============================================================================

; %1 is the name suffix of the function
; %2, %3 are the interpolations to use
; %4, %5, %6, %7 are the first interpolation parameters
; %8, %9, %10, %11 are the second interpolation parameters
%MACRO CONVERSION_FRAME 11
   avs_mangle yv12toyuy2_frame_ %+ %1 %+ _isse

yv12toyuy2_frame_ %+ %1 %+ _isse:
   AVS_SET_STACK     4, esi, edi, ebp, ebx
   
   .pDst             PARAM(1)
   .nDstPitch        PARAM(2)
   .pSrcY            PARAM(3)
   .nSrcYPitch       PARAM(4)
   .pSrcU            PARAM(5)
   .nSrcUPitch       PARAM(6)
   .pSrcV            PARAM(7)
   .nSrcVPitch       PARAM(8)
   .nWidth           PARAM(9)
   .nHeight          PARAM(10)
   
   .nWidthDiv2       PSTACK(0)
   .nLoopCounter     PSTACK(1)
   .nSrcUPitch2      PSTACK(2)
   .nSrcVPitch2      PSTACK(3)
   
   mov         ecx, [esp + .nHeight]
   sub         ecx, 2                        ; first & last line are processed differently
   mov         [esp + .nLoopCounter], ecx
   
   mov         ecx, [esp + .nSrcUPitch]
   add         ecx, ecx
   mov         [esp + .nSrcUPitch2], ecx

   mov         ecx, [esp + .nSrcVPitch]
   add         ecx, ecx
   mov         [esp + .nSrcVPitch2], ecx
   
   mov         ecx, [esp + .nWidth]
   shr         ecx, 1
   mov         [esp + .nWidthDiv2], ecx
   
   mov         edi, [esp + .pDst]
   mov         esi, [esp + .pSrcY]
   mov         ebx, [esp + .pSrcU]
   mov         ebp, [esp + .pSrcV]
   
   pxor        mm7, mm7
   
   ; first line, special --> chroma copy
   CONVERSION_LINE esi, ebx, ebp, ebx, ebp, CONVERSION_COPY_FIRST, eax, esp + .nWidthDiv2, 1, 0, 0, 0, edi
   add         esi, [esp + .nSrcYPitch]
   add         edi, [esp + .nDstPitch]
   mov         ecx, ebx
   mov         edx, ebp
   add         ecx, [esp + .nSrcUPitch]
   add         edx, [esp + .nSrcVPitch]
   
.loopy   

   CONVERSION_LINE esi, ebx, ebp, ecx, edx, %2, eax, esp + .nWidthDiv2, %4, %5, %6, %7, edi
   add         esi, [esp + .nSrcYPitch]
   add         edi, [esp + .nDstPitch]

   CONVERSION_LINE esi, ebx, ebp, ecx, edx, %3, eax, esp + .nWidthDiv2, %8, %9, %10, %11, edi
   add         esi, [esp + .nSrcYPitch]
   add         edi, [esp + .nDstPitch]
   add         ecx, [esp + .nSrcUPitch]
   add         edx, [esp + .nSrcVPitch]
   add         ebx, [esp + .nSrcUPitch]
   add         ebp, [esp + .nSrcVPitch]
   
   mov         eax, [esp + .nLoopCounter]
   sub         eax, 2
   mov         [esp + .nLoopCounter], eax
   
   jnz         .loopy
      
   ; last line, special --> chroma copy
   CONVERSION_LINE esi, ebx, ebp, ebx, ebp, CONVERSION_COPY_FIRST, eax, esp + .nWidthDiv2, 1, 0, 0, 0, edi

   AVS_REL_STACK
%ENDMACRO

;=============================================================================
; Code
;=============================================================================

SECTION .text:

CONVERSION_FRAME 1, CONVERSION_INTERPOLATION, CONVERSION_INTERPOLATION, 3, 1, 2, 2, 1, 3, 2, 2
CONVERSION_FRAME 2, CONVERSION_AVERAGE, CONVERSION_COPY_SECOND, 1, 1, 1, 1, 0, 1, 0, 0
CONVERSION_FRAME 3, CONVERSION_COPY_FIRST, CONVERSION_AVERAGE, 1, 0, 0, 0, 1, 1, 1, 1

CONVERSION_FRAME 1_top, CONVERSION_INTERPOLATION, CONVERSION_INTERPOLATION, 5, 3, 4, 3, 1, 7, 4, 3
CONVERSION_FRAME 1_bottom, CONVERSION_INTERPOLATION, CONVERSION_INTERPOLATION, 7, 1, 4, 3, 3, 5, 4, 3
